dnl !!!!!!!!!!!!! IMPORTANT TO BE DEFINED FIRST !!!!!!!!!!!!!
dnl #############################################################################################
dnl _NullStream         Represents the null stream, used for "swallowing" output
dnl #############################################################################################
define(`_NullStream',`-1')dnl
divert(_NullStream)dnl
changecom(`//', `')dnl comment characters are now c++ style instead of m4 style

dnl #############################################################################################
dnl _StdoutStream       Represents the stdout stream, used when resetting stream output after
dnl                     redirecting for a "divert"
dnl #############################################################################################
define(`_StdoutStream',`0')

dnl -----------------------------------------------
dnl   compiler for platform-common taxi
dnl -----------------------------------------------

dnl note convention: initial lowercase after _ indicates function
dnl                  initial uppercase after _ indicates variable

dnl ###########################
dnl   list of global variables
dnl ###########################

dnl _DriverName		The name of the driver module as it will be announced to the OS (eg fooDriver)
dnl			Automatically set when driver name is set
dnl _SmashedDriverName	The name of the driver, smashed to upper case
dnl			Automatically set by _setDriverName()
dnl _ErrorUndefined	An error message which remains in output if an invalid macro
dnl			is not overridden for some reason
define(`_ErrorUndefined', `/*** ERROR! "`$1'" should be overriden by an OS specific macro ***/
')
dnl _MajorNumber		The major number of the device
dnl _MinorNumber		The minor number of the device
define(`_DeclareTaxiRetval', `int TaxiRetval;')
define(`_TaxiRetval', `TaxiRetval')
define(`_MethodRetval', `MethodRetval')dnl

dnl ###########################
dnl   internal functions
dnl ###########################

dnl _headerIfndef(filename) takes the name of the header file, smashes case to upper,
dnl 	prefixes a leading '_' and translates any '.' to '_'
dnl	it outputs the standard c #ifndef-#define idiom for the top of a header file
define(`_headerIfndef',
`pushdef(`smashed', `translit($1, `a-z', `A-Z')')dnl
pushdef(`bashed', `translit(smashed, `.', `_')')dnl
`#'ifndef _`'bashed
`#'define _`'bashed
popdef(`bashed')dnl
popdef(`smashed')dnl')

dnl _headerEndif(filename) takes the name of the header file, smashes case to upper,
dnl 	prefixes a leading '_' and translates any '.' to '_'
dnl	it outputs the standard c #endif idiom for the end of a header file
define(`_headerEndif',
`pushdef(`smashed', `translit($1, `a-z', `A-Z')')dnl
pushdef(`bashed', `translit(smashed, `.', `_')')dnl
`#'endif /* _`'bashed */
popdef(`bashed')dnl
popdef(`smashed')dnl')

define(`_includeDriverHeader', ``#'include "_DriverName.h"')

dnl _ifndefDefine(macro) defines a symbol in the c pre-processor
dnl 	according to the traditional paradigm
define(`_ifndefDefine',
``#'ifndef $1
`#'define $1 $2
`#'endif')

define(`_starIndent', `patsubst($1,`@',` *    ')')

dnl _startHeader takes the name of the header file to be _DriverName.h
dnl 	_startHeader adds appropriate comments to top of file, and takes care of the whole
dnl     #ifndef #define buisness.
define(`_startHeader', `dnl
/*********************************************************************
 *  This is the beginning of c header file "_DriverName.h" for _Os
dnl # place author names in comment, if author names have been defined
ifdef(`_DriverAuthorNames', `dnl
 *  _DriverName written by:
_starIndent(_DriverAuthorNames)
')dnl
 *********************************************************************/

_headerIfndef(_DriverName.h)
')

dnl ###############################################################
dnl  _endHeader
dnl     > takes the name of the header file to be _DriverName.h
dnl 	  adds appropriate comments to end of file, and takes care 
dnl       of the whole #endif buisness.
define(`_endHeader',`dnl
_headerEndif(_DriverName.h)
/*********************************************************************
 *  This is the end of c header file "_DriverName.h" for _Os
 *********************************************************************/
')

dnl _startSourceFile takes the name of the header file to be _DriverName.h
dnl 	_startSourceFile adds appropriate comments to top of file.
define(`_startSourceFile', `
/**********************************************************************
 *  This is the beginning of c source file "_DriverName.c" for _Os
 *********************************************************************/
')

dnl _endSourceFile takes the name of the header file to be _DriverName.h
dnl 	_endSourceFile adds appropriate comments to top of file.
define(`_endSourceFile',`dnl
/*********************************************************************
 *  This is the end of c source file "_DriverName.c" for _Os
 *********************************************************************/
')

dnl _internalDefines defines internal (not exported to kernel) CPP symbols
dnl	who's NAMES should be common between all platforms (eg name of device, etc)
dnl	note that the VALUES of the symbols may be set by OS specific m4 files
define(`_internalDefines',
`define(`_MajorNumberSymbol', `_SmashedDriverName`'_MAJOR')dnl
_ifndefDefine(_MajorNumberSymbol,_MajorNumber)
define(`_MinorNumberSymbol', `_SmashedDriverName`'_MINOR')dnl
_ifndefDefine(_MinorNumberSymbol,_MinorNumber)
define(`_DriverNameSymbol', `_SmashedDriverName`'_NAME')dnl
_ifndefDefine(_DriverNameSymbol,_DriverName)')

dnl _addFuncPrototype(prototype)
dnl	adds the specified prototype to the list of prototypes in _FuncPrototypes
define(`_addFuncPrototype',dnl
`define(`_FuncPrototypes',dnl
defn(`_FuncPrototypes')
$1;
)')

dnl _addFuncDefinition(prototype)
dnl	adds the specified prototype to the list of prototypes in _FuncDefinitions
define(`_addFuncDefinition',dnl
`define(`_FuncDefinitions',dnl
defn(`_FuncDefinitions')
$1
)')

dnl ##########################################################
dnl   macros that should be overridden by OS specific values
dnl ##########################################################

dnl _osKernelDefines defines all CPP kernel compilation symbols that the target OS requires
define(`_osKernelDefines', `_ErrorUndefined(`_osKernelDefines')');
dnl _declareFopsPrototypes is a high level macro which spits out the prototypes for all
dnl	file operation functions which have been registered
define(`_declareFopsPrototypes', `_ErrorUndefined(`_declareFopsPrototypes')');
dnl _declareFopsStruct declares a file operations struct to be used in the driver's .c source
dnl 	the format of such a struct is obviously OS dependant.  under linux it comes from fs.h
dnl	each entry is a funtion pointer, which we implement with macros, which are initialized to NULL
dnl	until a particular file operation is registerd.
define(`_declareFopsStruct', `_ErrorUndefined(`_declareFopsStruct')')
define(`_osLicense', `_ErrorUndefined(`_osLicense')')
define(`_registerModule', `_ErrorUndefined(`_registerModule')')
define(`_Os', `_ErrorUndefined(`_Os')')

dnl #########################################
dnl   externally (user) available functions
dnl #########################################

dnl ####################################################
dnl     _name(name)
dnl	   > set name of driver, to be used throughout 
dnl          code to prevent namespace collisions on a 
dnl          primitive level
dnl ####################################################
define(`_name',
`define(`_DriverName', `$1')dnl
define(`_SmashedDriverName', `translit(_DriverName, `a-z', `A-Z')')')

dnl #########################################
dnl     _addAuthor(name)
dnl	   > adds specified name to list, one name per line
dnl #########################################
define(`_addAuthor',dnl
`define(`_DriverAuthorNames',dnl
defn(`_DriverAuthorNames')
@$1)')

dnl #########################################
dnl     _version(versionstring)
dnl	   > sets driver version
dnl #########################################
define(`_version', `define(`_DriverVersion', `$1')')

dnl #########################################
dnl     _description(description)
dnl	   > sets driver description
dnl #########################################
define(`_description', `define(`_DriverDescrip', `$1')')

dnl #########################################
dnl     _license(licenseType)
dnl	   > some OSs (linux) want to know whether or not the driver is
dnl        under the GPL.
dnl #########################################
define(`_license', `define(`_DriverLicenseType', `$1')')

dnl ################################################
dnl     _major(number)
dnl	   > sets the major number
dnl ################################################
define(`_major', `define(`_MajorNumber', `$1')')

dnl ################################################
dnl     _minor(number)
dnl	   > sets the minor number
dnl ################################################
define(`_minor', `define(`_MinorNumber', `$1')')

dnl ################################################
dnl     _return(retval)
dnl	   > returns value to caller of function
dnl ################################################
define(`_return', `return $1;')

dnl ###################################################
dnl     _markModuleUsage
dnl     _unmarkModuleUsage
dnl        > keeps count of module usage (semaphore?)
dnl ###################################################
define(`_markModuleUsage', `_ErrorUndefined(`_`'markModuleUsage()')')
define(`_unmarkModuleUsage', `_ErrorUndefined(`_`'unmarkModuleUsage()')')

dnl ################################################
dnl     _ifos(osName, conditional code)
dnl	   adds specified code to output if we are compiling a driver for specified os.
dnl ################################################
define(`_ifos', `ifelse(`$1', defn(`_Os'), `$2', `dnl')')

dnl ################################################
dnl     _cppMacro(symbol name, symbol value)
dnl		adds given macro to list of cpp macros
dnl ################################################
define(`_cppMacro',dnl
`define(`_CppMacros',dnl
defn(`_CppMacros')
`#'`'defin`'e $1 $2)')

dnl ################################################
dnl     _struct(name, def)
dnl		adds struct to list
dnl ################################################
define(`_struct',dnl
`define(`_Structs',dnl
defn(`_Structs')
struct $1 {$2})')

dnl ################################################
dnl     _func(proto, def)
dnl		adds function to list
dnl ################################################
define(`_func',dnl
`_addFuncPrototype($1)dnl
_addFuncDefinition($1 {
  _DeclareTaxiRetval
$2})dnl
')

dnl ########################################################################
dnl    The following is the high level taxi code for the actual generation
dnl    of the source files
dnl ########################################################################
define(`_compile', `dnl
dnl ---------------------------------------------------------------------------
dnl      this generates the c header file for the driver
dnl ---------------------------------------------------------------------------
_startHeader

_osIncludes

_OsDefines

_internalDefines

ifdef(`_CppMacros', _CppMacros) dnl misc user defined cpp macros

ifdef(`_Structs', _Structs) dnl misc user defined structs

_declareFopsPrototypes

ifdef(`_FuncPrototypes', _FuncPrototypes) dnl misc user defined functions

_endHeader

dnl ---------------------------------------------------------------------------
dnl      this generates the c source file for the driver
dnl ---------------------------------------------------------------------------
_startSourceFile

_includeDriverHeader

_osVars
_declareFopsStruct

_defineInitModule
_defineExitModule
_defineFops
ifdef(`_FuncDefinitions', _FuncDefinitions) dnl misc user defined functions

dnl add OS license macro, if any
_osLicense
_endSourceFile

')

define(`_driver', `dnl
divert(_NullStream)
$1
divert(_StdoutStream)dnl
_compile()
')dnl
